# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-core-wow64-l1-1-1.dll")
prototypes = \
    {
        #
        'GetSystemWow64Directory2A': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer", "uSize", "ImageFileMachineType"]),
        #
        'GetSystemWow64Directory2W': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer", "uSize", "ImageFileMachineType"]),
        #
        'Wow64SetThreadDefaultGuestMachine': SimTypeFunction([SimTypeShort(signed=False, label="UInt16")], SimTypeShort(signed=False, label="UInt16"), arg_names=["Machine"]),
    }

lib.set_prototypes(prototypes)



# pylint: disable=missing-class-docstring,no-self-use
from ailment.expression import Convert

from .base import PeepholeOptimizationExprBase


class RemoveNoopConversions(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Remove no-op conversions"
    expr_classes = (Convert,)

    def optimize(self, expr: Convert, **kwargs):
        if expr.from_bits == expr.to_bits:
            return expr.operand

        if isinstance(expr.operand, Convert):
            inner = expr.operand
            if (
                expr.to_bits == inner.from_bits
                and expr.from_bits == inner.to_bits
                and expr.is_signed == inner.is_signed
            ):
                return inner.operand

        return None



from typing import Tuple, Dict, Set, DefaultDict, Any, Optional
from collections import defaultdict

import claripy
import ailment
from angr.serializable import Serializable
from angr.knowledge_plugins.functions.function import Function
from .states import PropagatorVEXState, PropagatorAILState, PropagatorState


class PropagationModel(Serializable):
    """
    This class stores the propagation result that comes out of Propagator.
    """

    __slots__ = (
        "key",
        "node_iterations",
        "states",
        "input_states",
        "block_initial_reg_values",
        "replacements",
        "equivalence",
        # internals of the function graph visitor
        "graph_visitor",
        "_initial_state",
        "_function",
    )

    def __init__(
        self,
        prop_key: Tuple,
        node_iterations: Optional[DefaultDict[Any, int]] = None,
        states: Optional[Dict] = None,
        block_initial_reg_values: Optional[Dict] = None,
        replacements: Optional[DefaultDict[Any, Dict]] = None,
        equivalence: Optional[Set] = None,
        function: Optional[Function] = None,
        input_states: Optional[Dict] = None,
    ):
        self.key = prop_key
        self.node_iterations = node_iterations if node_iterations is not None else defaultdict(int)
        self.input_states = input_states if input_states is not None else {}
        self.states = states if states is not None else {}
        self.block_initial_reg_values = block_initial_reg_values if block_initial_reg_values is not None else {}
        self.replacements = replacements
        self.equivalence = equivalence if equivalence is not None else set()

        self.graph_visitor = None
        self._initial_state = None
        self._function = function

    def downsize(self):
        self.node_iterations = None
        self.block_initial_reg_values = None
        self.states = None
        self.input_states = None
        self.graph_visitor = None

    def block_beginning_state(self, block_addr) -> PropagatorState:
        if self._function is None:
            raise NotImplementedError()

        node = self._function.get_node(block_addr)
        preds = [self.states[pnode.addr] for pnode in self._function.graph.predecessors(node)]
        if not preds:
            if isinstance(node, ailment.Block):
                state = PropagatorAILState.initial_state(self._function.project, func_addr=self._function.addr)
            else:
                state = PropagatorVEXState.initial_state(self._function.project, func_addr=self._function.addr)
                state.store_register(state.arch.ip_offset, state.arch.bytes, claripy.BVV(block_addr, state.arch.bits))
        else:
            state, _ = preds[0].merge(*preds[1:])
        return state



from typing import Set

import ailment


class Goto:
    """
    Describe the existence of a goto (jump) statement. May have multiple gotos with the same address (targets
    will differ).
    """

    def __init__(self, src_addr, dst_addr, src_idx=None, dst_idx=None, src_ins_addr=None):
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.src_idx = src_idx
        self.dst_idx = dst_idx
        self.src_ins_addr = src_ins_addr

    def __hash__(self):
        return hash(f"{self.src_addr}{self.dst_addr}{self.src_idx}{self.dst_idx}")

    def __str__(self):
        if self.src_addr is None or self.dst_addr is None:
            return f"<Goto {self.__hash__()}>"

        src_idx_str = "" if self.src_idx is None else f".{self.src_idx}"
        dst_idx_str = "" if self.dst_idx is None else f".{self.dst_idx}"
        src_ins_addr_str = "" if self.src_ins_addr is None else f"{hex(self.src_ins_addr)}"
        return f"<Goto: [{hex(self.src_addr)}@{src_ins_addr_str}{src_idx_str}] -> {hex(self.dst_addr)}{dst_idx_str}>"

    def __repr__(self):
        return self.__str__()


class GotoManager:
    """
    Container class for all Gotos found in a function after decompilation structuring.
    This should be populated using GotoSimplifier.
    """

    def __init__(self, func, gotos=None):
        self.func = func
        self.gotos: Set[Goto] = gotos or set()

        self._gotos_by_addr = None

    def __str__(self):
        return f"<GotoManager: func[{hex(self.func.addr)}] {len(self.gotos)} gotos>"

    def __repr__(self):
        return self.__str__()

    def gotos_in_block(self, block: ailment.Block) -> Set[Goto]:
        gotos_found = set()
        for goto in self.gotos:
            if goto.src_addr == block.addr:
                gotos_found.add(goto)
            else:
                block_addrs = {stmt.ins_addr for stmt in block.statements if "ins_addr" in stmt.tags}
                if goto.src_ins_addr in block_addrs:
                    gotos_found.add(goto)

        return gotos_found

    def is_goto_edge(self, src: ailment.Block, dst: ailment.Block):
        src_gotos = self.gotos_in_block(src)
        for goto in src_gotos:
            if goto.dst_addr == dst.addr:
                return True
            else:
                block_addrs = {stmt.ins_addr for stmt in dst.statements if "ins_addr" in stmt.tags}
                if goto.dst_addr in block_addrs:
                    return True

        return False



import networkx  # pylint:disable=unused-import


class RemoveNodeNotice(Exception):
    pass


class AILGraphWalker:
    """
    Walks an AIL graph and optionally replaces each node with a new node.
    """

    def __init__(self, graph, handler, replace_nodes: bool = False):
        self.graph: networkx.DiGraph = graph
        self.handler = handler
        self._replace_nodes = replace_nodes

    def walk(self):
        for node in list(self.graph.nodes()):
            try:
                r = self.handler(node)
                remove = False
            except RemoveNodeNotice:
                # we need to remove this node
                r = None
                remove = True

            if self._replace_nodes:
                if remove:
                    self.graph.remove_node(node)
                elif r is not None and r is not node:
                    in_edges = list(self.graph.in_edges(node, data=True))
                    out_edges = list(self.graph.out_edges(node, data=True))

                    self.graph.remove_node(node)
                    self.graph.add_node(r)

                    for src, _, data in in_edges:
                        if src is node:
                            self.graph.add_edge(r, r, **data)
                        else:
                            self.graph.add_edge(src, r, **data)

                    for _, dst, data in out_edges:
                        if dst is node:
                            self.graph.add_edge(r, r, **data)
                        else:
                            self.graph.add_edge(r, dst, **data)



from .timing import timethis
from . import graph
from . import constants
from . import enums_conv
from . import lazy_import
from .env import is_pyinstaller


def looks_like_sql(s: str) -> bool:
    """
    Determine if string `s` looks like an SQL query.

    :param str s:   The string to detect.
    :return:        True if the string looks like an SQL, False otherwise.
    """

    sql_keywords = {"select", "update", "union", "delete", "from", "table", "insert", "into"}

    s = s.lower()
    for k in sql_keywords:
        if k in s:
            k_index = s.find(k)
            # what's before k? is it a whitespace if it's not empty?
            if k_index > 0:
                before = s[k_index - 1]
                if before not in " /;":
                    continue
            # what's after k? is it a whitespace?
            following = s[k_index + len(k) :]
            if not following or following.startswith(" "):
                return True

    return False



from typing import Optional, Type

from .dream import DreamStructurer
from .phoenix import PhoenixStructurer
from .recursive_structurer import RecursiveStructurer


STRUCTURER_CLASSES = {
    "dream": DreamStructurer,
    "phoenix": PhoenixStructurer,
}


def structurer_class_from_name(name: str) -> Optional[Type]:
    return STRUCTURER_CLASSES.get(name.lower(), None)



import pyvex

from .heavy import HeavyVEXMixin

from angr.state_plugins.sim_action import SimActionObject, SimActionData, SimActionExit, SimActionOperation
from angr import sim_options as o


class TrackActionsMixin(HeavyVEXMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__tmp_deps = {}

    __tls = ("__tmp_deps",)

    def _is_true(self, v):
        return super()._is_true(v[0])

    def _is_false(self, v):
        return super()._is_false(v[0])

    def _optimize_guarded_addr(self, addr, guard):
        addr, addr_deps = addr
        guard, _ = guard
        addr = super()._optimize_guarded_addr(addr, guard)
        return addr, addr_deps

    def handle_vex_block(self, irsb):
        self.__tmp_deps = {}
        super().handle_vex_block(irsb)

    def _handle_vex_const(self, const):
        return super()._handle_vex_const(const), frozenset()

    def _handle_vex_expr_GSPTR(self, expr):
        return super()._handle_vex_expr_GSPTR(expr), frozenset()

    def _handle_vex_expr_VECRET(self, expr):
        return super()._handle_vex_expr_VECRET(expr), frozenset()

    def _handle_vex_expr_Binder(self, expr):
        return super()._handle_vex_expr_Binder(expr), frozenset()

    def _instrument_vex_expr(self, result):
        return super()._instrument_vex_expr(result[0]), result[1]

    def _perform_vex_expr_Op(self, op, args):
        exprs, deps = zip(*args)
        result = super()._perform_vex_expr_Op(op, exprs)

        if o.TRACK_OP_ACTIONS in self.state.options:
            action_objects = [SimActionObject(arg, deps=dep, state=self.state) for arg, dep in args]
            r = SimActionOperation(self.state, op, action_objects, result)
            self.state.history.add_action(r)
            result_deps = frozenset((r,))
        else:
            result_deps = frozenset().union(*deps)
        return result, result_deps

    def _perform_vex_expr_ITE(self, *args):
        exprs, deps = zip(*args)
        combined_deps = frozenset().union(*deps)
        result = super()._perform_vex_expr_ITE(*exprs)
        return result, combined_deps

    # TODO for this and below: what if we made AUTO_DEPS work here?
    def _perform_vex_expr_CCall(self, func_name, ty, args, func=None):
        exprs, deps = zip(*args)
        combined_deps = frozenset().union(*deps)
        result = super()._perform_vex_expr_CCall(func_name, ty, exprs, func=None)
        return result, combined_deps

    def _perform_vex_stmt_Dirty_call(self, func_name, ty, args, func=None):
        exprs, deps = zip(*args) if args else ((), ())
        combined_deps = frozenset().union(*deps)
        result = super()._perform_vex_stmt_Dirty_call(func_name, ty, exprs, func=None)
        return result, combined_deps

    def _perform_vex_expr_RdTmp(self, tmp):
        result = super()._perform_vex_expr_RdTmp(tmp)

        # finish it and save the tmp reference
        if o.TRACK_TMP_ACTIONS in self.state.options:
            r = SimActionData(
                self.state,
                SimActionData.TMP,
                SimActionData.READ,
                tmp=tmp,
                size=self.irsb.tyenv.sizeof(tmp),
                data=result,
            )
            self.state.history.add_action(r)
            a = frozenset((r,))
        else:
            a = self.__tmp_deps.get(tmp, frozenset())
        return result, a

    def _perform_vex_expr_Get(self, offset_bundle, ty, **kwargs):
        offset, offset_deps = offset_bundle
        result = super()._perform_vex_expr_Get(offset, ty, **kwargs)

        if o.TRACK_REGISTER_ACTIONS in self.state.options:
            offset_ao = SimActionObject(offset, deps=offset_deps, state=self.state)
            r = SimActionData(
                self.state,
                self.state.registers.id,
                SimActionData.READ,
                addr=offset_ao,
                size=pyvex.get_type_size(ty),
                data=result,
            )
            self.state.history.add_action(r)
            a = frozenset((r,))
        else:
            a = frozenset()
        return result, a

    def _perform_vex_expr_Load(self, addr_bundle, ty, end, condition=None, **kwargs):
        addr, addr_deps = addr_bundle

        if condition is not None:
            condition, condition_deps = condition
        else:
            condition_deps = None

        result = super()._perform_vex_expr_Load(addr, ty, end, condition=condition, **kwargs)

        if o.TRACK_MEMORY_ACTIONS in self.state.options:
            addr_ao = SimActionObject(addr, deps=addr_deps, state=self.state)
            condition_ao = (
                SimActionObject(condition, deps=condition_deps, state=self.state) if condition is not None else None
            )
            r = SimActionData(
                self.state,
                self.state.memory.id,
                SimActionData.READ,
                addr=addr_ao,
                size=pyvex.get_type_size(ty),
                data=result,
                condition=condition_ao,
            )
            self.state.history.add_action(r)
            a = frozenset((r,))
        else:
            a = frozenset()
        return result, a

    def _perform_vex_stmt_LoadG_guard_condition(self, guard):
        return super()._perform_vex_stmt_LoadG_guard_condition(guard[0]), guard[1]

    def _perform_vex_stmt_StoreG_guard_condition(self, guard):
        return super()._perform_vex_stmt_StoreG_guard_condition(guard[0]), guard[1]

    # statements

    def _perform_vex_stmt_WrTmp(self, tmp, data_bundle, **kwargs):
        data, data_deps = data_bundle

        if o.TRACK_TMP_ACTIONS not in self.state.options:
            self.__tmp_deps[tmp] = data_deps
        super()._perform_vex_stmt_WrTmp(tmp, data, deps=data_deps)

    def _perform_vex_stmt_Put(self, offset_bundle, data_bundle, **kwargs):
        offset, offset_deps = offset_bundle
        data, data_deps = data_bundle
        # track the put
        if o.TRACK_REGISTER_ACTIONS in self.state.options:
            data_ao = SimActionObject(data, deps=data_deps, state=self.state)
            size_ao = SimActionObject(len(data))
            a = SimActionData(
                self.state, SimActionData.REG, SimActionData.WRITE, addr=offset, data=data_ao, size=size_ao
            )
            self.state.history.add_action(a)
        else:
            a = None

        super()._perform_vex_stmt_Put(offset, data, action=a, **kwargs)

    def _perform_vex_stmt_Store(self, addr_bundle, data_bundle, end, condition=None, **kwargs):
        addr, addr_deps = addr_bundle
        data, data_deps = data_bundle

        if condition is not None:
            condition, condition_deps = condition
        else:
            condition_deps = None

        # track the write
        if o.TRACK_MEMORY_ACTIONS in self.state.options and addr_deps is not None:
            data_ao = SimActionObject(data, deps=data_deps, state=self.state)
            addr_ao = SimActionObject(addr, deps=addr_deps, state=self.state)
            size_ao = SimActionObject(len(data))
            cond_ao = (
                SimActionObject(condition, deps=condition_deps, state=self.state)
                if condition_deps is not None
                else None
            )
            a = SimActionData(
                self.state,
                SimActionData.MEM,
                SimActionData.WRITE,
                data=data_ao,
                size=size_ao,
                addr=addr_ao,
                condition=cond_ao,
            )
            self.state.history.add_action(a)
        else:
            a = None

        super()._perform_vex_stmt_Store(addr, data, end, action=a, condition=condition, **kwargs)

    def _perform_vex_stmt_Exit(self, guard_bundle, target_bundle, jumpkind):
        guard, guard_deps = guard_bundle
        target, target_deps = target_bundle

        if o.TRACK_JMP_ACTIONS in self.state.options:
            guard_ao = SimActionObject(guard, deps=guard_deps, state=self.state)
            target_ao = SimActionObject(target, deps=target_deps, state=self.state)
            self.state.history.add_action(
                SimActionExit(self.state, target=target_ao, condition=guard_ao, exit_type=SimActionExit.CONDITIONAL)
            )

        super()._perform_vex_stmt_Exit(guard, target, jumpkind)

    def _perform_vex_defaultexit(self, target_bundle, jumpkind):
        if target_bundle is not None:
            target, target_deps = target_bundle

            if o.TRACK_JMP_ACTIONS in self.state.options:
                target_ao = SimActionObject(target, deps=target_deps, state=self.state)
                self.state.history.add_action(SimActionExit(self.state, target_ao, exit_type=SimActionExit.DEFAULT))
        else:
            target = None

        super()._perform_vex_defaultexit(target, jumpkind)



from collections import defaultdict
from itertools import count
import copy
import logging
import inspect

from .optimization_pass import OptimizationPassStage, StructuringOptimizationPass
from ..call_counter import AILBlockCallCounter

l = logging.getLogger(__name__)


class CrossJumpReverter(StructuringOptimizationPass):
    """
    This is an implementation to revert the compiler optimization Cross Jumping, and ISC optimization discussed
    in the USENIX 2024 paper SAILR. This optimization is somewhat aggressive and as such should be run last in your
    decompiler deoptimization chain. This deoptimization will take any goto it finds and attempt to duplicate its
    target block if its target only has one outgoing edge.

    There are some heuristics in place to prevent duplication everywhere. First, this deoptimization will only run
    a max of max_opt_iters times. Second, it will not duplicate a block with too many calls.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.DURING_REGION_IDENTIFICATION
    NAME = "Duplicate linear blocks with gotos"
    STRUCTURING = ["phoenix"]
    DESCRIPTION = inspect.cleandoc(__doc__).strip()

    def __init__(
        self,
        func,
        # internal parameters that should be used by Clinic
        node_idx_start: int = 0,
        # settings
        max_opt_iters: int = 3,
        max_call_duplications: int = 2,
        **kwargs,
    ):
        super().__init__(func, max_opt_iters=max_opt_iters, strictly_less_gotos=True, **kwargs)

        self.node_idx = count(start=node_idx_start)
        self._max_call_dup = max_call_duplications
        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):
        to_update = defaultdict(list)
        for node in self.out_graph.nodes:
            gotos = self._goto_manager.gotos_in_block(node)
            # TODO: support if-stmts
            if not gotos or len(gotos) >= 2:
                continue

            # only blocks that have a single outgoing goto are candidates
            # for duplicates
            goto = list(gotos)[0]
            for goto_target in self.out_graph.successors(node):
                if goto_target.addr == goto.dst_addr:
                    break
            else:
                goto_target = None

            # the target goto block should only have a single outgoing edge
            # this prevents duplication of conditions
            if goto_target is None or self.out_graph.out_degree(goto_target) != 1:
                continue

            # minimize the number of calls in the target block that can be duplicated
            # to prevent duplication of big blocks
            counter = AILBlockCallCounter()
            counter.walk(goto_target)
            if counter.calls > self._max_call_dup:
                continue

            # [goto_target] = (pred1, pred2, ...)
            to_update[goto_target].append(node)

        if not to_update:
            return False

        updates = False
        sorted_targets = sorted(to_update.items(), key=lambda x: x[0].addr)
        for goto_target, pred_to_update in sorted_targets:
            pred_to_update = sorted(pred_to_update, key=lambda x: x.addr)
            # do some sanity checks
            update_edges = [(pred, goto_target) for pred in pred_to_update]
            if not all(self.out_graph.has_edge(*edge) for edge in update_edges):
                continue

            current_preds = list(self.out_graph.predecessors(goto_target))
            delete_original = len(current_preds) == len(pred_to_update)

            # update the edges
            for src, goto_blk in update_edges:
                cp = copy.deepcopy(goto_blk)
                cp.idx = next(self.node_idx)
                self.out_graph.remove_edge(src, goto_blk)
                self.out_graph.add_edge(src, cp)

            updates = True
            if delete_original:
                self.out_graph.remove_node(goto_target)

        return updates



# pylint: disable=no-self-use,unused-private-member

import abc
import logging
import threading
from typing import Optional
import angr

from archinfo.arch_soot import SootAddressDescriptor

l = logging.getLogger(name=__name__)


class SimEngineBase:
    """
    Even more basey of a base class for SimEngine. Used as a base by mixins which want access to the project but for
    which having method `process` (contained in `SimEngine`) doesn't make sense
    """

    def __init__(self, project=None, **kwargs):
        if kwargs:
            raise TypeError("Unused initializer args: " + ", ".join(kwargs.keys()))
        self.project: Optional[angr.Project] = project
        self.state = None

    __tls = ("state",)

    def __getstate__(self):
        return (self.project,)

    def __setstate__(self, state):
        self.project = state[0]
        self.state = None

    def _is_true(self, v):
        return v is True

    def _is_false(self, v):
        return v is False


class SimEngine(SimEngineBase, metaclass=abc.ABCMeta):
    """
    A SimEngine is a class which understands how to perform execution on a state. This is a base class.
    """

    @abc.abstractmethod
    def process(self, state, **kwargs):
        """
        The main entry point for an engine. Should take a state and return a result.

        :param state:   The state to proceed from
        :return:        The result. Whatever you want ;)
        """


class TLSMixin:
    """
    Mix this class into any class that defines __tls to make all of the attributes named in that list into
    thread-local properties.

    MAGIC MAGIC MAGIC
    """

    def __new__(cls, *args, **kwargs):  # pylint:disable=unused-argument
        obj = super().__new__(cls)
        obj.__local = threading.local()
        return obj

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        for subcls in cls.mro():
            for attr in subcls.__dict__.get("_%s__tls" % subcls.__name__, ()):
                if attr.startswith("__"):
                    attr = f"_{subcls.__name__}{attr}"

                if hasattr(cls, attr):
                    if type(getattr(cls, attr, None)) is not TLSProperty:
                        raise Exception("Programming error: %s is both in __tls and __class__" % attr)
                else:
                    setattr(cls, attr, TLSProperty(attr))


class TLSProperty:  # pylint:disable=missing-class-docstring
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance._TLSMixin__local, self.name)

    def __set__(self, instance, value):
        setattr(instance._TLSMixin__local, self.name, value)

    def __delete__(self, instance):
        delattr(instance._TLSMixin__local, self.name)


class SuccessorsMixin(SimEngine):
    """
    A mixin for SimEngine which implements ``process`` to perform common operations related to symbolic execution
    and dispatches to a ``process_successors`` method to fill a SimSuccessors object with the results.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.successors: Optional[SimSuccessors] = None

    __tls = ("successors",)

    def process(self, state, *args, **kwargs):  # pylint:disable=unused-argument
        """
        Perform execution with a state.

        You should only override this method in a subclass in order to provide the correct method signature and
        docstring. You should override the ``_process`` method to do your actual execution.

        :param state:       The state with which to execute. This state will be copied before
                            modification.
        :param inline:      This is an inline execution. Do not bother copying the state.
        :param force_addr:  Force execution to pretend that we're working at this concrete address
        :returns:           A SimSuccessors object categorizing the execution's successor states
        """
        inline = kwargs.pop("inline", False)
        force_addr = kwargs.pop("force_addr", None)

        ip = state._ip
        addr = (
            (ip if isinstance(ip, SootAddressDescriptor) else state.solver.eval(ip))
            if force_addr is None
            else force_addr
        )

        # make a copy of the initial state for actual processing, if needed
        if not inline and o.COPY_STATES in state.options:
            new_state = state.copy()
        else:
            new_state = state
        # enforce this distinction
        old_state = state
        del state
        self.state = new_state

        # we have now officially begun the stepping process! now is where we "cycle" a state's
        # data - move the "present" into the "past" by pushing an entry on the history stack.
        # nuance: make sure to copy from the PREVIOUS state to the CURRENT one
        # to avoid creating a dead link in the history, messing up the statehierarchy
        new_state.register_plugin("history", old_state.history.make_child())
        new_state.history.recent_bbl_addrs.append(addr)
        if new_state.arch.unicorn_support:
            new_state.scratch.executed_pages_set = {addr & ~0xFFF}

        self.successors = SimSuccessors(addr, old_state)

        new_state._inspect(
            "engine_process", when=BP_BEFORE, sim_engine=self, sim_successors=self.successors, address=addr
        )
        self.successors = new_state._inspect_getattr("sim_successors", self.successors)
        try:
            self.process_successors(self.successors, **kwargs)
        except SimException as e:
            if o.EXCEPTION_HANDLING not in old_state.options:
                raise
            old_state.project.simos.handle_exception(self.successors, self, e)

        new_state._inspect("engine_process", when=BP_AFTER, sim_successors=self.successors, address=addr)
        self.successors = new_state._inspect_getattr("sim_successors", self.successors)

        # downsizing
        if new_state.supports_inspect:
            new_state.inspect.downsize()
        # if not TRACK, clear actions on OLD state
        # if o.TRACK_ACTION_HISTORY not in old_state.options:
        #    old_state.history.recent_events = []

        # fix up the descriptions...
        description = str(self.successors)
        l.info("Ticked state: %s", description)
        for succ in self.successors.all_successors:
            succ.history.recent_description = description
        for succ in self.successors.flat_successors:
            succ.history.recent_description = description

        return self.successors

    def process_successors(self, successors, **kwargs):
        """
        Implement this function to fill out the SimSuccessors object with the results of stepping state.

        In order to implement a model where multiple mixins can potentially handle a request, a mixin may implement
        this method and then perform a super() call if it wants to pass on handling to the next mixin.

        Keep in mind python's method resolution order when composing multiple classes implementing this method.
        In short: left-to-right, depth-first, but deferring any base classes which are shared by multiple subclasses
        (the merge point of a diamond pattern in the inheritance graph) until the last point where they would be
        encountered in this depth-first search. For example, if you have classes A, B(A), C(B), D(A), E(C, D), then the
        method resolution order will be E, C, B, D, A.

        :param state:           The state to manipulate
        :param successors:      The successors object to fill out
        :param kwargs:          Any extra arguments. Do not fail if you are passed unexpected arguments.
        """
        successors.processed = False  # mark failure


# pylint:disable=wrong-import-position
from .. import sim_options as o
from ..state_plugins.inspect import BP_BEFORE, BP_AFTER
from .successors import SimSuccessors
from ..errors import SimException



# pylint:disable=too-many-boolean-expressions
from typing import Union

from ailment.expression import Convert, BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ConstMullAShift(PeepholeOptimizationExprBase):
    """
    Convert expressions with right shifts into expressions with divisions.
    """

    __slots__ = ()

    NAME = "Conv(64->32, (N * a) >> M) => a / N1"
    expr_classes = (Convert, BinaryOp)

    def optimize(self, expr: Union[Convert, BinaryOp], **kwargs):
        r = None

        if isinstance(expr, Convert):
            if expr.from_bits == 64 and expr.to_bits == 32:
                r = self.optimize_binaryop(expr)

        elif isinstance(expr, BinaryOp):
            r = self.optimize_binaryop(expr)

        # keep size
        if r is not None and r.bits < expr.bits:
            r = Convert(expr.idx, r.bits, expr.bits, False, r, **expr.tags)

        return r

    def optimize_binaryop(self, expr: BinaryOp):
        if isinstance(expr, BinaryOp) and expr.op == "Shr" and isinstance(expr.operands[1], Const):
            # (N * a) >> M  ==>  a / N1
            inner = expr.operands[0]
            if isinstance(inner, BinaryOp) and inner.op == "Mull" and isinstance(inner.operands[0], Const):
                C = inner.operands[0].value
                X = inner.operands[1]
                V = expr.operands[1].value
                ndigits = 5 if V == 32 else 6
                divisor = self._check_divisor(pow(2, V), C, ndigits)
                if divisor is not None:
                    new_const = Const(None, None, divisor, X.bits)
                    new_div = BinaryOp(inner.idx, "Div", [X, new_const], inner.signed, **inner.tags)
                    return new_div

        elif isinstance(expr, BinaryOp) and expr.op in {"Add", "Sub"}:
            expr0, expr1 = expr.operands
            if (
                isinstance(expr0, BinaryOp)
                and expr0.op in {"Shr", "Sar"}
                and isinstance(expr0.operands[1], Const)
                and isinstance(expr1, BinaryOp)
                and expr1.op in {"Shr", "Sar"}
                and isinstance(expr1.operands[1], Const)
            ):
                if (
                    isinstance(expr0.operands[0], BinaryOp)
                    and expr0.operands[0].op in {"Mull", "Mul"}
                    and isinstance(expr0.operands[0].operands[1], Const)
                ):
                    a0 = expr0.operands[0].operands[0]
                    a1 = expr1.operands[0]
                    if a0 == a1:
                        # (a * x >> M1) +/- (a >> M2)  ==>  a / N
                        C = expr0.operands[0].operands[1].value
                        X = a0
                        V = expr0.operands[1].value
                        ndigits = 5 if V == 32 else 6
                        divisor = self._check_divisor(pow(2, V), C, ndigits)
                        if divisor is not None:
                            new_const = Const(None, None, divisor, X.bits)
                            new_div = BinaryOp(
                                expr0.operands[0].idx,
                                "Div",
                                [X, new_const],
                                expr0.operands[0].signed,
                                **expr0.operands[0].tags,
                            )
                            # we cannot drop the convert in this case
                            return new_div

        return None

    @staticmethod
    def _check_divisor(a, b, ndigits=6):
        divisor_1 = 1 + (a // b)
        divisor_2 = int(round(a / float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None



from ailment.statement import Assignment
from ailment.expression import BinaryOp, Const, Tmp

from .base import PeepholeOptimizationStmtBase


class RolRorRewriter(PeepholeOptimizationStmtBase):
    """
    Rewrites consecutive statements into ROL (rotate shift left) or ROR (rotate shift right) statements.
    """

    __slots__ = ()

    NAME = "ROL/ROR rewriter"
    stmt_classes = (Assignment,)

    def optimize(self, stmt: Assignment, stmt_idx: int = None, block=None, **kwargs):
        # Rol example:
        #    61 | t304 = Shr32(t301,0x19)
        #    62 | t306 = Shl32(t301,0x07)
        #    63 | t303 = Or32(t306,t304)
        #
        # Ror example:
        #    98 | 0x140002a06 | t453 = (Conv(64->32, r9<8>) << 0x11<8>)
        #    99 | 0x140002a06 | t455 = (Conv(64->32, r9<8>) >> 0xf<8>)
        #    100 | 0x140002a06 | t452 = (t455 | t453)
        if stmt_idx < 2:
            return None
        if not (isinstance(stmt.src, BinaryOp) and stmt.src.op == "Or"):
            return None

        op0, op1 = stmt.src.operands
        if not (isinstance(op0, Tmp) and isinstance(op1, Tmp)):
            return None
        # check the previous two instructions
        stmt_1 = block.statements[stmt_idx - 1]
        stmt_2 = block.statements[stmt_idx - 2]
        if not (isinstance(stmt_1, Assignment) and isinstance(stmt_1.src, BinaryOp)):
            return None
        if not (isinstance(stmt_2, Assignment) and isinstance(stmt_2.src, BinaryOp)):
            return None

        if not isinstance(stmt_1.dst, Tmp):
            return None
        if not isinstance(stmt_2.dst, Tmp):
            return None

        if {stmt_1.dst.tmp_idx, stmt_2.dst.tmp_idx} != {op0.tmp_idx, op1.tmp_idx}:
            return None

        stmt1_op0, stmt1_op1 = stmt_1.src.operands
        stmt2_op0, stmt2_op1 = stmt_2.src.operands

        if not (stmt1_op0.likes(stmt2_op0)):
            return None

        if not (isinstance(stmt1_op1, Const) and isinstance(stmt2_op1, Const)):
            return None

        if stmt_1.src.op == "Shl" and stmt_2.src.op == "Shr" and stmt1_op1.value + stmt2_op1.value == stmt.dst.bits:
            new_stmt = Assignment(
                stmt.idx,
                stmt.dst,
                BinaryOp(None, "Rol", [stmt1_op0, stmt1_op1], False, bits=stmt.dst.bits, **stmt_1.src.tags),
                **stmt.tags,
            )
            return new_stmt
        elif stmt_1.src.op == "Shr" and stmt_2.src.op == "Shl" and stmt1_op1.value + stmt2_op1.value == stmt.dst.bits:
            new_stmt = Assignment(
                stmt.idx,
                stmt.dst,
                BinaryOp(None, "Ror", [stmt1_op0, stmt1_op1], False, bits=stmt.dst.bits, **stmt_1.src.tags),
                **stmt.tags,
            )
            return new_stmt

        return None



from typing import Iterable, Dict, Optional

from . import MemoryMixin


class LabelMergerMixin(MemoryMixin):
    """
    A memory mixin for merging labels. Labels come from SimLabeledMemoryObject.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _merge_labels(self, labels: Iterable[Dict], **kwargs) -> Optional[Dict]:
        new_label = {}
        all_keys = set()
        for label in labels:
            all_keys.update(label.keys())
        for key in all_keys:
            v = ...
            for label in labels:
                if v is ...:
                    v = label.get(key)
                elif v != label.get(key):
                    v = None
            new_label[key] = v
        return new_label

    def copy(self, memo=None):
        copied = super().copy(memo)
        return copied
